"""
SPDX-License-Identifier: GPL-3.0-or-later
SPDX-FileCopyrightText: 2020 CERN
"""

import pytest
import time
from PyAdcLib import PyAdcConf, PyFmcAdc100m14b4ch, timeval
import numpy as np
import matplotlib.pyplot as plt

@pytest.fixture(scope="function")
def adc_simple_triangle(adc_simple):
    yield adc_simple

def digital_noise_plot(raw_data, index, chan_err):
    # convert signed data to mV (numpy array)
    plot_data = np.array(raw_data)
    # plot and store
    f1, a1 = plt.subplots()

    for ch in range(0,4):
        a1.plot(raw_data[ch::4], label='CH'+str(ch + 1) + (" WARN" if ch == chan_err else ""))

    a1.set_xlabel('sample number')
    a1.set_xticklabels(range(index, index + int(len(raw_data) / 4)))
    a1.set_ylabel('raw value')
    a1.legend()
    f1.savefig("/tmp/digital_noise_plot.png", dpi=300)

    plt.show()

def digital_noise_report(buf, index, chan_err):
    message = "Unexpected value on channel {:d}\n".format(chan_err)
    if index > 1:
        for chan in range(4):
            prev = buf.contents.get_sample(chan, index - 1)
            samp = buf.contents.get_sample(chan, index)
            diff = abs(samp - prev)
            message += "chan {:d}, sample {:d}: h{:x} = abs(h{:x} - h{:x}) = abs(b{:b} - b{:b})\n".format(chan,
                                                                                                          index,
                                                                                                          diff,
                                                                                                          prev,
                                                                                                          samp,
                                                                                                          prev,
                                                                                                          samp)
    subset = buf.contents.data[index * 4:(index+100) * 4]
    message += str(subset)
    digital_noise_plot(subset, index, chan_err)
    return message

class TestAdcRegressions(object):

    @pytest.mark.parametrize("vref", [PyFmcAdc100m14b4ch.ADC_CONF_100M14B4CHA_CHN_RANGE_100mV,
                                      #PyFmcAdc100m14b4ch.ADC_CONF_100M14B4CHA_CHN_RANGE_1V,
                                      #PyFmcAdc100m14b4ch.ADC_CONF_100M14B4CHA_CHN_RANGE_10V,
                                      ])
    @pytest.mark.parametrize("termination", [0, 1])
    @pytest.mark.parametrize("nshots", [1, 10])
    @pytest.mark.parametrize("post_samples", [2045, ])
    def test_adc_digital_noise(self, adc_simple_triangle, nshots, post_samples, vref, termination):
        """ With the v5 release we noticed some digital noise on the acquired signal.
        This test uses a sawtooth signal to check that the noise is between
        define boundaries. If it exceed, then we have a problem.
        A function generator must be configured to produce a triangle signal
        (simpler to process) on all channels. Expected configuration: 2Vpp, 800kHz"""
        for ch in range(4):
            conf = PyAdcConf(PyAdcConf.ADC_CONF_TYPE_CHN, ch)
            conf.value_set(PyAdcConf.ADC_CONF_CHN_TERMINATION, termination)
            conf.value_set(PyAdcConf.ADC_CONF_CHN_RANGE, vref)
            adc_simple_triangle.apply_config(conf, 0)

        conf = PyAdcConf(PyAdcConf.ADC_CONF_TYPE_ACQ, 0)
        conf.value_set(PyAdcConf.ADC_CONF_ACQ_PRE_SAMP, 0)
        conf.value_set(PyAdcConf.ADC_CONF_ACQ_POST_SAMP, post_samples)
        conf.value_set(PyAdcConf.ADC_CONF_ACQ_N_SHOTS, nshots)
        adc_simple_triangle.apply_config(conf, 0)

        buf = adc_simple_triangle.request_buffer(post_samples, None, 0)
        adc_simple_triangle.acq_start(PyFmcAdc100m14b4ch.ADC_F_FLUSH,
                                         timeval(0, 0))
        sleep_time = 0.5 + 1e-8 * post_samples
        for n in range(nshots):
            time.sleep(sleep_time)
            adc_simple_triangle.trigger_fire()

        for n in range(nshots):
            adc_simple_triangle.acq_poll(0, timeval(10, 0))
            adc_simple_triangle.fill_buffer(buf, 0, timeval(10, 0))
            assert buf.contents.nsamples == post_samples

            for chan in range(4):
                prev_s = None
                for i in range(post_samples):
                    sample = abs(buf.contents.get_sample(chan, i) >> 2)
                    if prev_s is not None:
                        # TODO improve future sample computarion (just rule of thumb number now)
                        assert abs(sample - prev_s) < 500, digital_noise_report(buf, i, chan)
                    prev_s = sample

        adc_simple_triangle.release_buffer(buf, None)
